__all__ = [
	'__author__',
	'__author_email__',
	'__copyright__',
	'__license__',
	'__summary__',
	'__title__',
	'__url__',
	'__version__',
	'__version_info__',
]

__title__ = 'google-music'
__summary__ = 'A Google Music API wrapper.'
__url__ = 'https://github.com/thebigmunch/google-music'

__version__ = '3.3.0'
__version_info__ = tuple(int(i) for i in __version__.split('.') if i.isdigit())

__author__ = 'thebigmunch'
__author_email__ = 'mail@thebigmunch.me'

__license__ = 'MIT'
__copyright__ = f'2018-2019 {__author__} <{__author_email__}>'
